/* Equalize 1.00 --- image filter plug-in for The Gimp image
 * manipulation program
 *
 * Copyright (C) 1996 Federico Mena Quintero
 *
 * You can contact me at federico@nuclecu.unam.mx
 * You can contact the original The Gimp authors at gimp@xcf.berkeley.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* This plug-in does an equalization of the image.  That is, it
 * modifies the image's histogram so that it is approximately
 * horizontal, so that all intensity values cover about the same
 * number of pixels in the image (i.e. the histogram is
 * well-balanced).
 *
 * @(GIMP)         = <plug-in equalize "Map/Equalize">
 * @(GIMP_DEP)     = <equalize.c>
 * @(GIMP_OBJ)     = <equalize.o>
 * @(GIMP_LIB)     = <c m>
 * @(GIMP_AUTHOR)  = <Federico Mena Quintero>
 * @(GIMP_EMAIL)   = <federico@nuclecu.unam.mx>
 * @(GIMP_DESC)    = <Equalize an image>
 * @(GIMP_VERSION) = <1.00>
 * @(GIMP_URL)     = <http://www.nuclecu.unam.mx/~federico/gimp/index.html>
 */


#include <stdio.h>
#include <stdlib.h>

#include "gimp.h"


/***** Prototypes *****/

static void do_equalize(Image source_imt, Image dest_img);
static void eq_histogram(long hist[3][256], unsigned char lut[3][256],
			 long channels, long width, long height);


/***** Variables *****/

static char *prog_name;


/***** Functions *****/

/*****/

int
main(int argc, char **argv)
{
	Image source, dest;

	/* Save program name */

	prog_name = argv[0];

	/* Initialize filter and continue if success */

	if (!gimp_init(argc, argv))
		return 0;

	source = gimp_get_input_image(0);

	if (source) {
		if ((gimp_image_type(source) == GRAY_IMAGE) ||
		    (gimp_image_type(source) == RGB_IMAGE)) {
			dest = gimp_get_output_image(0);

			if (dest) {
				gimp_init_progress("Equalize");
				do_equalize(source, dest);
				gimp_update_image(dest);
				gimp_free_image(dest);
			} /* if */
		} else
			gimp_message("Equalize: can only operate on grayscale or RGB images");

		gimp_free_image(source);
	} /* if */

	gimp_quit();
	return 0;
} /* main */


/*****/

static void
do_equalize(Image source_img, Image dest_img)
{
	long           width, height;
	long           channels, rowsiz;
	unsigned char *src_data, *dest_data;
	unsigned char *src_line, *dest_line;
	unsigned char *src, *dest;
	long           hist[3][256]; /* Histogram */
	unsigned char  lut[3][256];
	int            x1, y1, x2, y2;
	long           x, y;
	int            k;
	int            progress, max_progress;

	/* Get selection area */

	gimp_image_area(source_img, &x1, &y1, &x2, &y2);

	width    = gimp_image_width(source_img);
	height   = gimp_image_height(source_img);
	channels = gimp_image_channels(source_img);
	rowsiz   = width * channels;

	/* Get image data */

	src_data  = (unsigned char *) gimp_image_data(source_img)
		    + channels * (width * y1 + x1);
	dest_data = (unsigned char *) gimp_image_data(dest_img)
		    + channels * (width * y1 + x1);

	/* Initialize progress */

	progress     = 0;
	max_progress = (y2 - y1) * 2;

	/* Initialize histogram */

	for (x = 0; x < 256; x++)
		for (k = 0; k < channels; k++)
			hist[k][x] = 0;

	/* Build histogram */

	src_line  = src_data;

	for (y = y1; y < y2; y++) {
		src  = src_line;
		
		for (x = x1; x < x2; x++)
			for (k = 0; k < channels; k++)
				hist[k][*src++]++;

		src_line += rowsiz;

		/* Update progress */

		if (progress++ % 64 == 0)
			gimp_do_progress(progress, max_progress);
	} /* for */

	/* Build equalization LUT */
	
	eq_histogram(hist, lut, channels, width, height);

	/* Substitute */

	src_line  = src_data;
	dest_line = dest_data;

	for (y = y1; y < y2; y++) {
		src  = src_line;
		dest = dest_line;

		for (x = x1; x < x2; x++)
			for (k = 0; k < channels; k++)
				*dest++ = lut[k][*src++];

		src_line  += rowsiz;
		dest_line += rowsiz;

		/* Update progress */

		if (progress++ % 64 == 0)
			gimp_do_progress(progress, max_progress);
	} /* for */
} /* do_equalize */


/*****/

static void
eq_histogram(long hist[3][256], unsigned char lut[3][256],
	     long channels, long width, long height)
{
	int    i, k, j;
	int    part[3][257]; /* Partition */
	double pixels_per_value;
	double desired;

	/* Calculate partial sums */

	for (k = 0; k < channels; k++)
		for (i = 1; i < 256; i++)
			hist[k][i] += hist[k][i - 1];

	/* Find partition points */

	pixels_per_value = (double) (width * height) / 256;

	for (k = 0; k < channels; k++) {
		/* First and last points in partition */

		part[k][0]   = 0;
		part[k][256] = 256;

		/* Find intermediate points */

		j = 0;
		
		for (i = 1; i < 256; i++) {
			desired = i * pixels_per_value;
			
			while (hist[k][j + 1] <= desired)
				j++;

			/* Nearest sum */

			if ((desired - hist[k][j]) < (hist[k][j + 1] - desired))
				part[k][i] = j;
			else
				part[k][i] = j + 1;
		} /* for */
	} /* for */

	/* Create equalization LUT */

	for (k = 0; k < channels; k++)
		for (j = 0; j < 256; j++) {
			i = 0;

			while (part[k][i + 1] <= j)
				i++;

			lut[k][j] = i;
		} /* for */
} /* eq_histogram */
